@@ -24,6 +24,7 @@ int *xp = NULL;
 double *W = NULL;
 double *V = NULL;
 double *gam = NULL;
+double *dof = NULL;
 
 int prexx;
 double *xbar = NULL;
@@ -40,8 +41,10 @@ double *E = NULL;
 double *Z = NULL;
 double vsum;
 
+double *l1fixedcost = NULL;
+
 unsigned int npass,nrw;
-double l1pen;
+double ntimeslam;
 
 double ysum,ybar;
 double *H = NULL;
@@ -66,24 +69,32 @@ void gamlr_cleanup(){
   dirty = 0;
 }
 
+/* The gradient descent move for given direction */
+double Bmove(int j)
+{
+  double dbet, pen, ghb;
+  if(H[j]==0) return -B[j];
 
-// calculates degrees of freedom, as well as
-// other gradient dependent statistics.
-double dof(int s, double *lam, double L){
+  pen = l1fixedcost[j]*nd;
+  if(W[j] > 0.0) pen += ntimeslam*W[j]*omega[j];
+  ghb = (G[j] - H[j]*B[j]);
+  if(fabs(ghb) < pen) dbet = -B[j];
+  else dbet = -(G[j]-sign(ghb)*pen)/H[j];
+
+  return dbet;
+}
+
+void donullgrad(void){
+  for(int j=0; j<p; j++)
+    if( (W[j]>0.0) & isfinite(W[j]) & (B[j]==0.0) ){
+      ag0[j] = fabs(G[j])/W[j] - l1fixedcost[j]*nd;
+      if(ag0[j]<0.0) ag0[j] = 0.0; 
+    }
+}
+
+double getdf(double L){
   int j;
-  
-  //calculate absolute grads
-  for(j=0; j<p; j++)
-    if(isfinite(W[j]) & (B[j]==0.0))
-      ag0[j] = fabs(G[j])/W[j];
-
-  // initialization  
-  if(s==0){
-    if(!isfinite(lam[0]))  
-      lam[0] = dmax(ag0,p)/nd;
-  }
-
-  double df = 1.0;
+  double dfs = 1.0;
 
   // penalized bit
   double shape,phi;
@@ -92,33 +103,13 @@ double dof(int s, double *lam, double L){
   for(j=0; j<p; j++)
     if(isfinite(W[j])){
       if( (gam[j]==0.0) | (W[j]==0.0) ){  
-        if( (B[j]!=0.0) ) df ++;
-      } else{ // gamma lasso
-        shape = lam[s]*nd/gam[j];
-        df += pgamma(ag0[j], shape/phi, phi*gam[j], 1, 0); 
+        if( (B[j]!=0.0) ) dfs++;
+      } else{ 
+        shape = ntimeslam/gam[j];
+        dfs += pgamma(ag0[j], shape/phi, phi*gam[j], 1, 0); 
       }
-    }
-
-  return df;
-}
-
-/* The gradient descent move for given direction */
-double Bmove(int j)
-{
-  double dbet;
-  if(H[j]==0) return -B[j];
-
-  // unpenalized
-  if(W[j]==0.0) dbet = -G[j]/H[j]; 
-  else{
-    // penalty is lam[s]*nd*W[j]*omega[j]*fabs(B[j])
-    double pen, ghb;
-    pen = l1pen*W[j]*omega[j];
-    ghb = (G[j] - H[j]*B[j]);
-    if(fabs(ghb) < pen) dbet = -B[j];
-    else dbet = -(G[j]-sign(ghb)*pen)/H[j];
   }
-  return dbet;
+  return(dfs);
 }
 
 void doxbar(void){
@@ -166,7 +157,7 @@ int cdsolve(double tol, int M, int RW)
   double dbet,imove,Bdiff;
 
   // initialize
-  dopen = isfinite(l1pen);
+  dopen = isfinite(ntimeslam);
   Bdiff = INFINITY;
   exitstat = 0;
   dozero = 1;
@@ -275,12 +266,13 @@ int cdsolve(double tol, int M, int RW)
             int *nlam, // length of the path
             double *delta, // path stepsize
             double *gamvec,  // gamma in the GL paper
+            double *fixedcost, // additional fixed cost
             double *thresh,  // cd convergence
             int *maxit, // cd max iterations 
             int *maxrw, // max irls reweights
             double *lambda, // output lambda
             double *deviance, // output deviance
-            double *df, // output df
+            double *dofvec, // output dof
             double *alpha,  // output intercepts
             double *beta, // output coefficients
             int *exits, // exit status.  0 is normal, 1 warn, 2 break path
@@ -289,6 +281,7 @@ int cdsolve(double tol, int M, int RW)
   dirty = 1; // flag to say the function has been called
   // time stamp for periodic R interaction
   time_t itime = time(NULL);  
+  l1fixedcost = fixedcost;
 
   /** Build global variables **/
   fam = *famid;
@@ -341,6 +334,7 @@ int cdsolve(double tol, int M, int RW)
   G = new_dzero(p);
   ag0 = new_dzero(p);
   gam = gamvec;
+  dof = dofvec;
 
   // some local variables
   double Lold, NLLHD, NLsat;
@@ -372,10 +366,7 @@ int cdsolve(double tol, int M, int RW)
       for(int j=0; j<p; j++) dograd(j);
   }
 
-  l1pen = INFINITY;
-  Lold = INFINITY;
   NLLHD =  nllhd(n, A, E, Y, V);
-
   if(*verb)
     speak("*** n=%d observations and p=%d covariates ***\n", n,p);
 
@@ -385,7 +376,7 @@ int cdsolve(double tol, int M, int RW)
     // deflate the penalty
     if(s>0)
       lambda[s] = lambda[s-1]*(*delta);
-    l1pen = lambda[s]*nd;
+    ntimeslam = lambda[s]*nd;
 
     // run descent
     exits[s] = cdsolve(*thresh,maxit[s],maxrw[s]);
@@ -394,10 +385,16 @@ int cdsolve(double tol, int M, int RW)
     maxit[s] = npass;
     maxrw[s] = nrw;
     Lold = NLLHD;
-    if( (N>0) | (s==0) ) 
+    if( (s==0) | (N>0) ) 
       NLLHD =  nllhd(n, A, E, Y, V);
     deviance[s] = 2.0*(NLLHD - NLsat);
-    df[s] = dof(s, lambda, NLLHD);
+
+    donullgrad();
+    if( (s==0) & !isfinite(lambda[0])){ 
+      lambda[0] = dmax(ag0,p)/nd;
+      ntimeslam = lambda[0]*nd; }
+
+    dof[s] = getdf(NLLHD); 
     alpha[s] = A;
     copy_dvec(&beta[s*p],B,p);
 
@@ -407,8 +404,10 @@ int cdsolve(double tol, int M, int RW)
     for(int j=0; j<p; j++) 
       if(gam[j]>0.0){
         if(isfinite(gam[j])){
-          if( (W[j]>0.0) & isfinite(W[j]) )
-            omega[j] = 1.0/(1.0+gam[j]*fabs(B[j])); } 
+          if( (W[j]>0.0) & isfinite(W[j]) ){
+            omega[j] = 1.0/(1.0+gam[j]*fabs(B[j])); 
+          }
+        } 
         else if(B[j]!=0.0) omega[j] = 0.0; 
       }
       
@@ -418,11 +417,15 @@ int cdsolve(double tol, int M, int RW)
           s+1, lambda[s], deviance[s], npass);
 
     // exit checks
+    if(deviance[s]!=deviance[s]){
+      exits[s] = 2;
+      shout("Warning: NaN deviance.  ");
+    }
     if(deviance[s]<0.0){
       exits[s] = 2;
       shout("Warning: negative deviance.  ");
     }
-    if(df[s] >= nd){
+    if(dof[s] >= nd){
       exits[s] = 2;
       shout("Warning: saturated model.  "); 
     }
